import smtpd, base64, secure_smtpd, asynchat
from asyncore import ExitNow
from smtpd import NEWLINE, EMPTYSTRING

class SMTPChannel(smtpd.SMTPChannel):
    
    def __init__(self, smtp_server, newsocket, fromaddr, require_authentication=False, credential_validator=None, debug=False, map=None):
        smtpd.SMTPChannel.__init__(self, smtp_server, newsocket, fromaddr)
        asynchat.async_chat.__init__(self, newsocket, map=map)
        self.require_authentication = require_authentication
        self.authenticating = False
        self.authenticated = False
        self.username = None
        self.password = None
        self.debug = debug
        self.credential_validator = credential_validator
    
    def smtp_QUIT(self, arg):
        self.push('221 Bye')
        self.close_when_done()
        raise ExitNow()
        
    def collect_incoming_data(self, data):
        self.__line.append(data)
    
    def smtp_EHLO(self, arg):
        if not arg:
            self.push('501 Syntax: HELO hostname')
            return
        if self.__greeting:
            self.push('503 Duplicate HELO/EHLO')
        else:
            self.push('250-%s Hello %s' %  (self.__fqdn, arg))
            self.push('250-AUTH LOGIN')
            self.push('250 EHLO')
    
    def smtp_AUTH(self, arg):
        if 'LOGIN' in arg:
            self.authenticating = True
            split_args = arg.split(' ')
            
            # Some implmentations of 'LOGIN' seem to provide the username
            # along with the 'LOGIN' stanza, hence both situations are
            # handled.
            if len(split_args) == 2:
                self.username = base64.b64decode( arg.split(' ')[1] )
                self.push('334 ' + base64.b64encode('Username'))
            else:
                self.push('334 ' + base64.b64encode('Username'))
                
        elif not self.username:
            self.username = base64.b64decode( arg )
            self.push('334 ' + base64.b64encode('Password'))
        else:
            self.authenticating = False
            self.password = base64.b64decode(arg)
            if self.credential_validator and self.credential_validator.validate(self.username, self.password):
                self.authenticated = True
                self.push('235 Authentication successful.')
            else:
                self.push('454 Temporary authentication failure.')
                raise ExitNow()
    
    # This code is taken directly from the underlying smtpd.SMTPChannel
    # support for AUTH is added.
    def found_terminator(self):
        line = EMPTYSTRING.join(self.__line)
        
        if self.debug:
            secure_smtpd.logger.info('found_terminator(): data: %s' % repr(line))
            
        self.__line = []
        if self.__state == self.COMMAND:
            if not line:
                self.push('500 Error: bad syntax')
                return
            method = None
            i = line.find(' ')
            
            if self.authenticating:
                # If we are in an authenticating state, call the
                # method smtp_AUTH.
                arg = line.strip()
                command = 'AUTH'
            elif i < 0:
                command = line.upper()
                arg = None
            else:
                command = line[:i].upper()
                arg = line[i+1:].strip()
            
            # White list of operations that are allowed prior to AUTH.
            if not command in ['AUTH', 'EHLO', 'HELO', 'NOOP', 'RSET', 'QUIT']:
                if self.require_authentication and not self.authenticated:
                    self.push('530 Authentication required')
                    
            method = getattr(self, 'smtp_' + command, None)
            if not method:
                self.push('502 Error: command "%s" not implemented' % command)
                return
            method(arg)
            return
        else:
            if self.__state != self.DATA:
                self.push('451 Internal confusion')
                return
            # Remove extraneous carriage returns and de-transparency according
            # to RFC 821, Section 4.5.2.
            data = []
            for text in line.split('\r\n'):
                if text and text[0] == '.':
                    data.append(text[1:])
                else:
                    data.append(text)
            self.__data = NEWLINE.join(data)
            status = self.__server.process_message(
                self.__peer,
                self.__mailfrom,
                self.__rcpttos,
                self.__data
            )
            self.__rcpttos = []
            self.__mailfrom = None
            self.__state = self.COMMAND
            self.set_terminator('\r\n')
            if not status:
                self.push('250 Ok')
            else:
                self.push(status)



import secure_smtpd
import ssl, smtpd, time, asyncore, os
from smtp_channel import SMTPChannel
from multiprocessing import Process
from asyncore import ExitNow

class SMTPServer(smtpd.SMTPServer):
    
    KILL_SIGNAL = 9
    MAXIMUM_EXECUTION_TIME = 2
        
    def __init__(self, localaddr, remoteaddr, ssl=False, certfile=None, keyfile=None, ssl_version=ssl.PROTOCOL_SSLv23, require_authentication=False, credential_validator=None, debug=False):
        smtpd.SMTPServer.__init__(self, localaddr, remoteaddr)
        self.debug = debug
        self.certfile = certfile
        self.keyfile = keyfile
        self.ssl_version = ssl_version
        self.subprocesses = []
        self.require_authentication = require_authentication
        self.credential_validator = credential_validator
        self.ssl = ssl
        
    def handle_accept(self):
        if self.debug:
            secure_smtpd.logger.info('handle_accept(): called.')
        
        process = Process(target=self._accept_subprocess, args=[])
        self.subprocesses.append({
            'process': process,
            'start_time': time.time()
        })
        process.start()
        
        self._terminate_expired_subprocesses()
        
    def _terminate_expired_subprocesses(self):
        current_time = time.time()
        
        def alive(process_dict):
            if (current_time - process_dict['start_time']) > self.MAXIMUM_EXECUTION_TIME:
                return False
            return process_dict['process'].is_alive()
            
        self.subprocesses = filter(alive, self.subprocesses)
        
    def _accept_subprocess(self):
        try:
            pair = self.accept()
            map = {}
            
            if self.debug:
                secure_smtpd.logger.info('_accept_subprocess(): smtp connection accepted within subprocess.')
            
            if pair is not None:
                newsocket, fromaddr = pair
                if self.ssl:
                    newsocket = ssl.wrap_socket(
                        newsocket,
                        server_side=True,
                        certfile=self.certfile,
                        keyfile=self.keyfile,
                        ssl_version=self.ssl_version,
                    )
                channel = SMTPChannel(
                    self,
                    newsocket,
                    fromaddr,
                    require_authentication=self.require_authentication,
                    credential_validator=self.credential_validator,
                    debug=self.debug,
                    map=map
                )
                
                if self.debug:
                    secure_smtpd.logger.info('_accept_subprocess(): starting asyncore within subprocess.')
                
                asyncore.loop(map=map)
                secure_smtpd.logger.error('_accept_subprocess(): asyncore loop exited.')
        except ExitNow:
            if self.debug:
                secure_smtpd.logger.info('_accept_subprocess(): smtp channel terminated asyncore.')
            os.kill( os.getpid(), self.KILL_SIGNAL )


import logging
from logging.handlers import RotatingFileHandler
from smtp_server import SMTPServer
from fake_credential_validator import FakeCredentialValidator

def _create_logger():
    log_name = '/var/log/secure_smtpd.log'
    logger = logging.getLogger( log_name )
    logger.setLevel(logging.DEBUG)
    handler = RotatingFileHandler(
        log_name, 
        maxBytes=10485760,
        backupCount=3
    )
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger

logger = _create_logger()



#!/usr/bin/env python
#from distutils.core import setup
from setuptools import setup, find_packages

setup(
    name="secure-smtpd",
    version="1.0.0",
    description="Adds support for SSL, AUTH, and other goodies, to Petri Lehtinen's SMTPD library.",
    author="Benjamin Coe",
    author_email="bencoe@gmail.com",
    url="https://github.com/bcoe/secure-smtpd",
    packages = find_packages(),
    install_requires = [],
    tests_require=[
        'nose'
    ]
)


import secure_smtpd

# Implment this interface with an actual
# methodlogy for validating credentials, e.g.,
# lookup credentials for a usaer in Redis.
class FakeCredentialValidator(object):
    
    def validate(self, username, password):
        
        secure_smtpd.logger.warn('FakeCredentialValidator: you should replace this with an actual implementation of a credential validator.')
        
        if username == 'bcoe' and password == 'foobar':
            return True
        return False